from s3.settings import settings

from s3.s3_connector import s3, S3

try:
    from s3 import asyncio

    __all__ = ["s3", "S3", "asyncio", "settings"]
except ImportError:
    __all__ = ["s3", "S3", "settings"]
    pass



try:
    from s3.asyncio.s3_connector import async_s3, AsyncS3, AsyncS3Instance

    __all__ = ["async_s3", "AsyncS3", "AsyncS3Instance"]
except ImportError:
    raise ImportError("Async functions disabled. Please install extra 'async'")



import asyncio
import logging
from io import BufferedReader, BufferedIOBase

import aioboto3
from async_lru import alru_cache
from typing import Optional, Union

import botocore.exceptions
from aiofiles.threadpool.binary import AsyncBufferedReader, AsyncBufferedIOBase

from s3.settings import settings


class AsyncS3Instance():
    def __init__(self, client, logger, bucket):
        self._client = client
        self._logger = logger
        self._bucket = bucket

    @property
    def s3client(self):
        return self._client

    async def has_file(self, object_name: str):
        try:
            await self._client.head_object(Bucket=self._bucket, Key=object_name)
            return True
        except botocore.exceptions.ClientError:
            return False

    @staticmethod
    async def _universal_zero_seek(file: Union[AsyncBufferedIOBase, BufferedIOBase]):
        result = file.seek(0)
        if asyncio.iscoroutine(file):
            result = await result
        return result

    async def upload_file(self, file: Union[AsyncBufferedReader, BufferedReader], object_name: str):
        await self._universal_zero_seek(file)
        await self._client.upload_fileobj(file, self._bucket, object_name)
        await self._universal_zero_seek(file)

    async def download_file(self, file: Union[AsyncBufferedIOBase, BufferedIOBase], object_name: str):
        if not await self.has_file(object_name):
            raise FileNotFoundError(f"File '{object_name}' not found in '{self._bucket}' bucket")
        await self._universal_zero_seek(file)
        await self._client.download_fileobj(self._bucket, object_name, file)
        await self._universal_zero_seek(file)

    async def delete_file(self, object_name: str):
        if await self.has_file(object_name):
            await self._client.delete_object(Bucket=self._bucket, Key=object_name)
        else:
            self._logger.warning(f"Deleting a nonexistent file '{object_name}' in '{self._bucket}' bucket")

    async def create_bucket(self):
        try:
            try:
                await self._client.head_bucket(Bucket=self._bucket)
                return
            except botocore.exceptions.ClientError:
                pass
            await self._client.create_bucket(Bucket=self._bucket)
        except botocore.exceptions.ClientError as e:
            raise e
        except Exception as e:
            self._logger.warning(f"Could not create bucket: {type(e)}, {str(e)}")
            raise e

    async def _generate_link(self, method: str, object_name: str, content_type: Optional[str], expiration: int) -> str:
        if not method.endswith("_object"):
            raise NotImplementedError(f"Method '{method}' not implemented")
        params = {'Bucket': self._bucket, 'Key': object_name}
        if content_type is not None:
            params['ContentType'] = content_type
        return await self._client.generate_presigned_url(
            ClientMethod=method,
            Params=params,
            ExpiresIn=expiration
        )

    async def generate_get_link(self, object_name, content_type: Optional[str] = None, expiration: int = 3600):
        return await self._generate_link('get_object', object_name, content_type, expiration)

    async def generate_put_link(self, object_name, content_type: Optional[str] = None, expiration: int = 3600):
        return await self._generate_link('put_object', object_name, content_type, expiration)

    async def __aenter__(self):
        self._client = await self._client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        ret = await self._client.__aexit__(exc_type, exc_val, exc_tb)
        return ret

    async def close(self):
        await self._client.close()


class AsyncS3:
    def __init__(self, bucket: Optional[str] = None,
                 access_key_id: Optional[str] = None,
                 secret_access_key: Optional[str] = None,
                 region_name: Optional[str] = None,
                 endpoint_url: Optional[str] = None):
        self._bucket = bucket if bucket is not None else settings.AWS_BUCKET
        self._logger = logging.getLogger(f"S3 on bucket '{self._bucket}'")
        if access_key_id is None:
            self._session = aioboto3.session.Session(aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
                                                     aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
                                                     region_name=settings.AWS_REGION)
            self._endpoint_url = settings.AWS_HOST
        else:
            self._session = aioboto3.session.Session(aws_access_key_id=access_key_id,
                                                     aws_secret_access_key=secret_access_key,
                                                     region_name=region_name)
            self._endpoint_url = endpoint_url

    @property
    def bucket(self):
        return self._bucket

    def __call__(self):
        s3client = self._session.client(service_name='s3', endpoint_url=self._endpoint_url)
        return AsyncS3Instance(s3client, self._logger, self._bucket)


@alru_cache(10)
async def async_s3(bucket: Optional[str] = None,
                   access_key_id: Optional[str] = None,
                   secret_access_key: Optional[str] = None,
                   region_name: Optional[str] = None,
                   endpoint_url: Optional[str] = None):
    s3 = AsyncS3(bucket=bucket,
                 access_key_id=access_key_id,
                 secret_access_key=secret_access_key,
                 region_name=region_name,
                 endpoint_url=endpoint_url)
    async with s3() as client:
        await client.create_bucket()
    return s3



import logging
from functools import lru_cache
from io import BufferedRandom
from typing import Optional

import boto3

import botocore.exceptions

from s3.settings import settings


class S3:
    def __init__(self, bucket: Optional[str] = None,
                 access_key_id: Optional[str] = None,
                 secret_access_key: Optional[str] = None,
                 region_name: Optional[str] = None,
                 endpoint_url: Optional[str] = None):
        self._bucket = bucket if bucket is not None else settings.AWS_BUCKET
        self._logger = logging.getLogger(f"S3 on bucket '{self._bucket}'")
        if access_key_id is None:
            self._session = boto3.session.Session(aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
                                                  aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
                                                  region_name=settings.AWS_REGION)
            self._s3client = self._session.client(service_name='s3', endpoint_url=settings.AWS_HOST)
        else:
            self._session = boto3.session.Session(aws_access_key_id=access_key_id,
                                                  aws_secret_access_key=secret_access_key,
                                                  region_name=region_name)
            self._s3client = self._session.client(service_name='s3', endpoint_url=endpoint_url)
        self.create_bucket()

    @property
    def bucket(self):
        return self._bucket

    @property
    def s3client(self):
        return self._s3client

    def has_file(self, object_name: str):
        try:
            self._s3client.head_object(Bucket=self._bucket, Key=object_name)
            return True
        except botocore.exceptions.ClientError:
            return False

    def upload_file(self, file: BufferedRandom, object_name: str):
        self._s3client.upload_fileobj(file, self._bucket, object_name)

    def download_file(self, file: BufferedRandom, object_name: str):
        if not self.has_file(object_name):
            raise FileNotFoundError(f"File '{object_name}' not found in '{self._bucket}' bucket")
        file.seek(0)
        self._s3client.download_fileobj(self._bucket, object_name, file)
        file.seek(0)

    def delete_file(self, object_name: str):
        if self.has_file(object_name):
            self._s3client.delete_object(Bucket=self._bucket, Key=object_name)
        else:
            self._logger.warning(f"Deleting a nonexistent file '{object_name}' in '{self._bucket}' bucket")

    def create_bucket(self):
        try:
            try:
                self._s3client.head_bucket(Bucket=self._bucket)
                return
            except botocore.exceptions.ClientError:
                pass
            self._s3client.create_bucket(Bucket=self._bucket)
        except botocore.exceptions.ClientError as e:
            raise e
        except Exception as e:
            self._logger.warning(f"Could not create bucket: {type(e)}, {str(e)}")
            raise e

    def _generate_link(self, method: str, object_name: str, content_type: Optional[str], expiration: int) -> str:
        if not method.endswith("_object"):
            raise NotImplementedError(f"Method '{method}' not implemented")
        params = {'Bucket': self._bucket, 'Key': object_name}
        if content_type is not None:
            params['ContentType'] = content_type
        return self._s3client.generate_presigned_url(
            ClientMethod=method,
            Params=params,
            ExpiresIn=expiration
        )

    def generate_get_link(self, object_name, content_type: Optional[str] = None, expiration: int = 3600):
        return self._generate_link('get_object', object_name, content_type, expiration)

    def generate_put_link(self, object_name, content_type: Optional[str] = None, expiration: int = 3600):
        return self._generate_link('put_object', object_name, content_type, expiration)


@lru_cache(10)
def s3(bucket: Optional[str] = None,
       access_key_id: Optional[str] = None,
       secret_access_key: Optional[str] = None,
       region_name: Optional[str] = None,
       endpoint_url: Optional[str] = None):
    return S3(bucket=bucket,
              access_key_id=access_key_id,
              secret_access_key=secret_access_key,
              region_name=region_name,
              endpoint_url=endpoint_url)



from pydantic_settings import BaseSettings
from typing import Optional


class Settings(BaseSettings):
    AWS_HOST: Optional[str] = None
    AWS_ACCESS_KEY_ID: Optional[str] = None
    AWS_SECRET_ACCESS_KEY: Optional[str] = None
    AWS_REGION: Optional[str] = None
    AWS_BUCKET: Optional[str] = None


settings = Settings()
